/* SPDX-License-Identifier: LGPL-2.1+ */

#ifndef __LXC_REXEC_H
#define __LXC_REXEC_H

#include "config.h"

#include "compiler.h"

__hidden extern int lxc_rexec(const char *memfd_name);

#endif /* __LXC_REXEC_H */
